/* The MazeGraphics, when instantiated, creates a brand
   new maze in which robots may be placed.
   Use this file to instantiate mazes and draw robots, gold and walls
   on them. This file does everything, screen-wise, in this respect! */

/* Abstract Programmer's Interface: here's how to use this class! :
   create a new maze by declaring a variable of type MazeClass:
   MazeGraphics maze = new MazeGraphics(<thisframe>, <width in cells>,
                                  <height in cells>, <anchor_x>,
                                  <anchor_y>)
    use anchor to specify the exact x,y pixel coordinate in the frame
    of the top left corner of the desired maze.
   use addWall(<cell_x>,<cell_y>,<dir>) to specify new walls. <dir> is
    a number from 0 to 3, with 0 being straight up (north) and 3 being
    to the right (east)
   use addGold() to add a gold piece. Gold that goes in a single cell
    is specified using addGold(<cell_x>,<cell_y>). Gold that sits on the
    border between two cells is specified via those two cells. (4 param's)
   use addRobot(<cell_x>,<cell_y>,<dir>,<numgolds>) to specify the position
    and direction of the robot. <cell_x>,<cell_y> are integers giving the
    coordinates in the maze, <dir> is 0 for up, 1 for left, 2 for down, and 3 for right;
    <numgolds> is 0,1,or 2 specifying the number of gold pieces in the robot.
   use addSmoothRobot(<cell_x>,<cell_y>,<angle>,<numgolds>) to place
    a robot in a non-grid-aligned position and orientation.  This is useful for
    displaying where within a cell the robot thinks it is.  <cell_x>,<cell_y>
    are doubles for this call, and the robot is centered in a cell when both are
    integer values.  <angle> is different for this call: it is a double representing
    the angle, in radians, that the robot is at, with zero facing to the right, and
    increasing counterclockwise.
   refresh() or update() is used to redraw the maze. You must specify this call in
    your window-redraw event handler: public void paint(Graphics g).
   removeWall works the obvious way, as does removeGold, removeRobot, and
    removeSmoothRobot.
   also, don't forget removeAllRobots, removeAllGolds and removeAllWalls
*/


import java.awt.*;
import java.util.*;
//import java.lang.reflect.*;
import java.io.*;

public class MazeGraphics extends java.lang.Object
{
    transient Frame parent_frame;
    int anchor_x, anchor_y; // top left anchor point for maze
    float magnification;
    int width, height; // number of cells width & height of world //
    int[][][] walls;
    Vector golds; // vector of int[] for each gold position
    Vector ports; // vector of int[] for each port position
    Vector robots; // vector of int[] for each robot position //
    int cellsize; // # of pixels width/height of a single cell //
    int wall_thickness; // thickness of a wall

    Vector smooth_robots; // vector of SmoothPoses for continuous robot poses

    private class SmoothRobotPose
    {
        double x, y, angle;
        int num_golds;
    }

    public MazeGraphics(Frame parent_frame, int width, int height,
                     int x, int y)
    {
        this.parent_frame = parent_frame;
        this.anchor_x = x;
        this.anchor_y = y;
        this.magnification = (float)1.0;
        this.cellsize = 40;
        this.wall_thickness = 4;
        this.width = width; this.height = height;
        // by default, set walls to have no walls everywhere //
        this.initializeWalls(this.width, this.height);
        this.robots = new Vector();
        this.golds = new Vector();
        this.ports = new Vector();
        this.smooth_robots = new Vector();
        try {
            Thread.sleep(1000);
        } catch (InterruptedException e) {
        }
        refresh();
    }

    public void initializeWalls(int width, int height)
    {
        this.walls = new int[width][height][4];
        for (int i=0; i < width; i++)
         for (int j=0; j < height; j++)
           for (int k=0; k < 4; k++)
             this.walls[i][j][k] = 0;
        // after zeroing everything, make the border 1's!!! //
        for (int i=0; i < width; i++) {
            this.walls[i][0][2] = 1;
            this.walls[i][height - 1][0] = 1;
        } // endfor
        for (int j=0; j < height; j++) {
            this.walls[0][j][1] = 1;
            this.walls[width - 1][j][3] = 1;
        } // endfor
    }

    // moves the maze to a new position in the world.
    public boolean moveMaze(int new_x, int new_y)
    {
        // undraw everything! //
        // get background color and make a filled rec
        this.anchor_x = new_x;
        this.anchor_y = new_y;
        // now redraw everything! //
        this.refresh();
        return(true);
    } // moveMaze() //

    public boolean drawWalls()
    {
        for (int i=0; i < this.width; i++)
            for (int j=0; j < this.height; j++) {
                this.cellWallspec(i,j,this.walls[i][j]);
            } // end for (int j... //
        return(true);
    } // drawWalls() //

    public boolean drawRobots()
    {
        for (int i=0; i < this.robots.size(); i++) {
            int[] robotpos = (int[])this.robots.elementAt(i);
            drawRobot(robotpos,i);
        } // end for //

        for (int i=0; i < this.robots.size(); i++) {
            int[] robotpos = (int[])this.robots.elementAt(i);
            drawRobotLine(robotpos,i);
        } // end for //

        for (int i = 0; i < this.smooth_robots.size(); i++) {
            drawSmoothRobot((SmoothRobotPose)this.smooth_robots.elementAt(i));
        }

        return(true);
    } // drawRobots() //

    public boolean drawRobots(int x, int y)
    {
        for (int i=0; i < this.robots.size(); i++) {
            int[] robotpos = (int[])this.robots.elementAt(i);
            if ((robotpos[0] == x) && (robotpos[1] == y))
                drawRobot(robotpos,i);
        } // end for //

        for (int i=0; i < this.robots.size(); i++) {
            int[] robotpos = (int[])this.robots.elementAt(i);
            if ((robotpos[0] == x) && (robotpos[1] == y))
                drawRobotLine(robotpos,i);
        } // end for //

        return(true);
    } // drawRobots(x,y) //

    public boolean drawPorts()
    {
        for (int i=0; i < this.ports.size(); i++) {
            int[] portspec = (int[])this.ports.elementAt(i);
            this.drawPort(portspec[0],portspec[1]);
        } // endfor
        return(true);
    } //drawPorts() //

    public boolean drawGolds()
    {
        for (int i=0; i < this.golds.size(); i++) {
            int[] goldspec = (int[])this.golds.elementAt(i);
            if (goldspec.length == 2) {
                this.drawGold(goldspec[0],goldspec[1]);
            } else if (goldspec.length == 4) {
                this.drawGold(goldspec[0],goldspec[1],
                              goldspec[2],goldspec[3]);
            } else return(false);
        } // endfor
        return(true);
    } //drawGolds() //
    public boolean drawGolds(int x, int y)
    {
        for (int i=0; i < this.golds.size(); i++) {
            int[] goldspec = (int[])this.golds.elementAt(i);
            if (goldspec.length == 2) {
                if ((goldspec[0] == x) && (goldspec[1] == y))
                    this.drawGold(goldspec[0],goldspec[1]);
            } else if (goldspec.length == 4) {
                if (((goldspec[0] == x) && (goldspec[1] == y)) ||
                    ((goldspec[2] == x) && (goldspec[3] == y)))
                    this.drawGold(goldspec[0],goldspec[1],
                                  goldspec[2],goldspec[3]);
            } else return(false);
        } // endfor
        return(true);
    } //drawGolds(x,y) //


    //takes a cell (x,y) and a direction and adds a wall there//
    public boolean addWall(int cell_x, int cell_y, int dir)
    {
        this.walls[cell_x][cell_y][dir] = 1;
        // add the symmetric cases if such a cell exists //
        int sym_x, sym_y, sym_dir;
        sym_dir = (dir + 2) % 4;
        if (dir == 0) {
            sym_x = cell_x; sym_y = cell_y + 1;
        } else if (dir == 1) {
            sym_x = cell_x - 1; sym_y = cell_y;
        } else if (dir == 2) {
            sym_x = cell_x; sym_y = cell_y - 1;
        } else {
            sym_x = cell_x + 1; sym_y = cell_y;
        }
        if ((sym_x >= 0) && (sym_y >= 0) &&
            (sym_x < this.width) && (sym_y < this.height))
                this.walls[sym_x][sym_y][sym_dir] = 1;
        drawWall(cell_x, cell_y, dir);
        return(true);
    } // addWall() //

    public boolean removeAllWalls()
    {
        this.initializeWalls(this.width, this.height);
        return(true);
    } // removeAllWalls() //

    public boolean removeWall(int cell_x, int cell_y, int dir)
    {
        // do not remove any border walls! new -- 10/12/98 illah //
        if ((cell_x == 0) && (dir == 1)) return(false);
        if ((cell_y == 0) && (dir == 2)) return(false);
        if ((cell_x == (this.width - 1)) && (dir == 3)) return(false);
        if ((cell_y == (this.height - 1)) && (dir == 0)) return(false);

        this.walls[cell_x][cell_y][dir] = 0;
        // add the symmetric cases if such a cell exists //
        int sym_x, sym_y, sym_dir;
        sym_dir = (dir + 2) % 4;
        if (dir == 0) {
            sym_x = cell_x; sym_y = cell_y + 1;
        } else if (dir == 1) {
            sym_x = cell_x - 1; sym_y = cell_y;
        } else if (dir == 2) {
            sym_x = cell_x; sym_y = cell_y - 1;
        } else {
            sym_x = cell_x + 1; sym_y = cell_y;
        }
        if ((sym_x >= 0) && (sym_y >= 0) &&
            (sym_x < this.width) && (sym_y < this.height))
                this.walls[sym_x][sym_y][sym_dir] = 0;
        eraseWall(cell_x, cell_y, dir);
        return(true);

    } // removeWall() //

    // The display is automatically refreshed at the end of this call.
    public boolean addRobot(int x, int y, int dir, int numgolds)
    {
        int[] robotpos = new int[4];
        robotpos[0] = x; robotpos[1]=y; robotpos[2]=dir;
        robotpos[3] = numgolds;
        int Counter = 0;
        for (int Index = 0; Index < robots.size(); Index++) {
            int [] tempRobot = (int [])robots.elementAt(Index);
            if (tempRobot[0] == x && tempRobot[1] == y)
                Counter++;
        }
        this.robots.addElement(robotpos);
        refresh();
        return(true);
    } // addRobot() //

    // Add a smoothly-positioned robot to the display.  When x and y have integer
    // values, the robot is centered in a cell, the same as with addRobot().
    // Unlike addRobot(), the angle parameter is 0 for East (right), and increases
    // counterclockwise in radians.  North is Math.PI/2, West is Math.PI, etc.
    public void addSmoothRobot(double x, double y, double angle, int num_golds)
    {
        SmoothRobotPose pose = new SmoothRobotPose();
        pose.x = x;
        pose.y = y;
        pose.angle = angle;
        pose.num_golds = num_golds;
        this.smooth_robots.addElement(pose);
        drawSmoothRobot(pose);
    }

    // Remove a robot smoothly-positioned robot from the display.  All parameters
    // must match *exactly* those of a previously added robot in order to delete it.
    // There is no epsilon position or angle distance tolerance.
    // The display is automatically refreshed at the end of this call.
    public void removeSmoothRobot(double x, double y, double angle, int num_golds)
    {
        boolean removed = false;

        for(int i = 0; i < this.smooth_robots.size(); i++) {
            SmoothRobotPose pose = (SmoothRobotPose) this.smooth_robots.elementAt(i);
            if(pose.x == x &&
               pose.y == y &&
               pose.angle == angle &&
               pose.num_golds == num_golds) {
                this.smooth_robots.removeElementAt(i);
                removed = true;
            }
        }
        if(removed)
            refresh();
    }

    // The display is automatically refreshed at the end of this call.
    public boolean removeAllRobots()
    {
        this.robots = new Vector();
        this.smooth_robots = new Vector();
        refresh();
        return(true);
    } // removeAllRobots()
    public boolean removeRobot(int x, int y, int dir, int golds)
    {
        boolean removed = false;

        for (int i=0; i < this.robots.size(); i++) {
            int[] robotpos = (int [])(this.robots.elementAt(i));
            if ((robotpos[0] == x) && (robotpos[1] == y) &&
                (robotpos[2] == dir) && (robotpos[3] == golds)) {
                    this.robots.removeElementAt(i);
                    removed = true;
            } // endif
        } // endfor
        if(removed) {
            refresh();
            return(true);
        }
        else {
            return(false);
        }
    } // removeRobot()

    public boolean addPort(int x, int y)
    {
        int[] portpos = new int[2];
        portpos[0] = x; portpos[1] = y;
        this.ports.addElement(portpos);
        drawPort(x, y);
        return(true);
    } // addPort() //

    public boolean addGold(int x, int y)
    {
        int[] goldpos = new int[2];
        goldpos[0] = x; goldpos[1] = y;
        this.golds.addElement(goldpos);
        drawGold(x, y);
        return(true);
    } // addGold() //
    public boolean addGold(int x1, int y1, int x2, int y2)
    {
        int[] goldpos = new int[4];
        goldpos[0]=x1; goldpos[1]=y1; goldpos[2]=x2; goldpos[3]=y2;
        this.golds.addElement(goldpos);
        drawGold(x1, y1, x2, y2);
        return(true);
    } // addGold(2)()

    public boolean removeAllGolds()
    {
        this.golds = new Vector();
        refresh();
        return(true);
    } // removeAllGolds() //

    public boolean removePort(int x, int y)
    {
        return(this.removePort(x,y,this.ports));
    } // removePort() //
    //this one takes a vector
    public boolean removePort(int x, int y, Vector daports)
    {
        for (int i=0; i < daports.size(); i++) {
            int[] portpos = (int[])(daports.elementAt(i));
            if ((portpos[0] == x) && (portpos[1] == y)) {
                erasePortPos(x, y);
                daports.removeElementAt(i);
                return(true);
            } //endif
        } //endfor
        return(false);
    } // removePort() //

    public boolean removeGold(int x, int y)
    {
        return(this.removeGold(x,y,this.golds));
    } // removeGold() //
    //this one takes a vector
    public boolean removeGold(int x, int y, Vector dagolds)
    {
        for (int i=0; i < dagolds.size(); i++) {
            int[] goldpos = (int[])(dagolds.elementAt(i));
            if ((goldpos.length == 2) &&
                (goldpos[0] == x) && (goldpos[1] == y)) {
                eraseGoldPos(x, y);
                dagolds.removeElementAt(i);
                return(true);
            } //endif
        } //endfor
        return(false);
    } // removeGold() //

    public boolean hasGold(int x, int y)
    {
        for (int i=0; i < this.golds.size(); i++) {
            int[] goldpos = (int[])(this.golds.elementAt(i));
            if ((goldpos.length == 2) &&
                (goldpos[0] == x) && (goldpos[1] == y)) {
                    return(true);
            } // endif
        } // endfor
        return(false);
    } // hasGold() //

    public boolean removeGold(int x1, int y1, int x2, int y2)
    {
        return(this.removeGold(x1,y1,x2,y2,this.golds));
    } // removeGold() //

    public boolean removeGold(int x1, int y1, int x2, int y2, Vector dagolds)
    {
        for (int i=0; i < dagolds.size(); i++) {
            int[] goldpos = (int[])(dagolds.elementAt(i));
            if ((goldpos.length == 4) &&
                (((goldpos[0] == x1) && (goldpos[1] == y1) &&
                 (goldpos[2] == x2) && (goldpos[3] == y2)) ||
                 ((goldpos[0] == x2) && (goldpos[1] == y2) &&
                 (goldpos[2] == x1) && (goldpos[3] == y1))))
                 {
                dagolds.removeElementAt(i);
                eraseGoldPos(xCell2Pixel(x1), xCell2Pixel(y1), xCell2Pixel(x2), xCell2Pixel(y2));
                if (x1 == x2 && y1 == y2-1)
                    eraseWall(x1, y1, 0);
                else if (x1 == x2 && y1 == y2+1)
                    eraseWall(x1, y1, 2);
                else if (y1 == y2 && x1 == x2 -1)
                    eraseWall(x1, y1, 3);
                else if (y1 == y2 && x1 == x2 + 1)
                    eraseWall(x1, y1, 1);
                return(true);
            } //endif
        } //endfor
        return(false);
    } // removeGold()2 //

    public int[] findOtherGoldCell(int x, int y)
    {
        return(this.findOtherGoldCell(x,y,this.golds));
    } //
    public int[] findOtherGoldCell(int x, int y, Vector dagolds)
    {
        for (int i=0; i < dagolds.size(); i++) {
            int[] goldpos = (int[])(dagolds.elementAt(i));
            if (goldpos.length == 4) {
                if ((goldpos[0] == x) && (goldpos[1] == y)) {
                    int[] temp = new int[2];
                    temp[0] = goldpos[2]; temp[1] = goldpos[3];
                    return(temp);
                }
                if ((goldpos[2] == x) && (goldpos[3] == y)) {
                    int[] temp = new int[2];
                    temp[0] = goldpos[0]; temp[1] = goldpos[1];
                    return(temp);
                }
            } // end if (goldpos.length == 4
        } // endfor
        return((int[])null);
    } // findOtherGoldCell() //

    // this searches for any accessible piece of gold (from here) and
    // removes it.
    public boolean removeAnyGold(int x, int y, Vector golds)
    {
        if (this.removeGold(x,y,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x-1,y,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x+1,y,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x,y-1,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x,y+1,golds)) {return(true);}
        return(false);
    } // removeAnyGold() //
    // this version does *not* remove gold at the bays, though.
    // this is for robot "control"
    public boolean removeAnyGoldRobot(int x, int y, Vector golds)
    {
        if ((x > 0) && (x < this.width-1) &&
            (this.removeGold(x,y,golds))) {return(true);}
        else
        if (this.removeGold(x,y,x-1,y,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x+1,y,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x,y-1,golds)) {return(true);}
        else
        if (this.removeGold(x,y,x,y+1,golds)) {return(true);}
        return(false);
    } // removeAnyGoldRobot() //


    // takes an x,y cell and a wallspec and draws the wallspec for the cell
    public boolean cellWallspec(int x, int y, int[] wallspec)
    {
        for (int Index = 0; Index < 4; Index ++) {
            if (wallspec[Index] == 1)
                drawWall(x, y, Index);
        }
        return(true);
    } // cellWallSpec() //

    public boolean drawWall(int x, int y, int dir)
    {
        switch(dir) {
        case 0:
            drawHorizontalLine(this.xCell2Pixel(x),
                               this.xCell2Pixel(x+1),
                               this.yCell2Pixel(y),
                               this.wall_thickness);
            break;
        // if (wallspec[0]) //
        case 2:
            drawHorizontalLine(this.xCell2Pixel(x),
                               this.xCell2Pixel(x+1),
                               this.yCell2Pixel(y-1),
                               this.wall_thickness);
            break;
        // if (wallspec[2]) //
        case 1:
            drawVerticalLine(this.xCell2Pixel(x),
                             this.yCell2Pixel(y),
                             this.yCell2Pixel(y-1),
                             this.wall_thickness);
            break;
        // if (wallspec[1]) //
        case 3:
            drawVerticalLine(this.xCell2Pixel(x+1),
                             this.yCell2Pixel(y),
                             this.yCell2Pixel(y-1),
                             this.wall_thickness);
            break;
        // if (wallspec[3]) //
        }
        return true;
    }

    public boolean eraseWall(int x, int y, int dir)
    {
        switch(dir) {
        case 0:
            eraseHorizontalLine(this.xCell2Pixel(x),
                                this.xCell2Pixel(x+1),
                                this.yCell2Pixel(y),
                                this.wall_thickness);
            drawHorizontalLine(this.xCell2Pixel(x),
                                this.xCell2Pixel(x+1),
                                this.yCell2Pixel(y),
                                1);
            break;
        // if (wallspec[0]) //
        case 2:
            eraseHorizontalLine(this.xCell2Pixel(x),
                                this.xCell2Pixel(x+1),
                                this.yCell2Pixel(y-1),
                                this.wall_thickness);
            drawHorizontalLine(this.xCell2Pixel(x),
                                this.xCell2Pixel(x+1),
                                this.yCell2Pixel(y-1),
                                1);
            break;
        // if (wallspec[2]) //
        case 1:
            eraseVerticalLine(this.xCell2Pixel(x),
                              this.yCell2Pixel(y),
                              this.yCell2Pixel(y-1),
                              this.wall_thickness);
            drawVerticalLine(this.xCell2Pixel(x),
                              this.yCell2Pixel(y),
                              this.yCell2Pixel(y-1),
                              1);
            break;
        // if (wallspec[1]) //
        case 3:
            eraseVerticalLine(this.xCell2Pixel(x+1),
                              this.yCell2Pixel(y),
                              this.yCell2Pixel(y-1),
                              this.wall_thickness);
            drawVerticalLine(this.xCell2Pixel(x+1),
                              this.yCell2Pixel(y),
                              this.yCell2Pixel(y-1),
                              1);
            break;
        } // if (wallspec[3]) //
        return true;
    }

    // takes the cell,cell x,y position & direction of the robot & draws
    public boolean drawRobot(int[] robotdesc, int robotnumber)
    {
        int cell_x= robotdesc[0];
        int cell_y=robotdesc[1] ;
        int dir=robotdesc[2];
        int numgolds = robotdesc[3];
        int x,y; // drawRobotCircle needs upper left corner of cell block
        x = this.xCell2Pixel(cell_x);
        y = this.yCell2Pixel(cell_y);
        return(this.drawRobotCircle(x,y,dir,numgolds,robotnumber));
    } // drawRobot() //

    public boolean drawRobotLine(int[] robotdesc, int robotnumber)
    {
        int cell_x= robotdesc[0];
        int cell_y=robotdesc[1] ;
        int dir=robotdesc[2];
        int numgolds = robotdesc[3];
        int x,y; // drawRobotLine needs upper left corner of cell block
        x = this.xCell2Pixel(cell_x);
        y = this.yCell2Pixel(cell_y);
        return(this.drawRobotLine(x,y,dir,numgolds,robotnumber));
    } // drawRobotLines() //

    // takes the cell,cell xy position of a port and draws it
    public boolean drawPort(int cell_x, int cell_y)
    {
        int x,y; // drawGoldPos needs upper x,y coordinates //
        x = this.xCell2Pixel(cell_x);
        y = this.yCell2Pixel(cell_y);
        return(this.drawPortPos(x,y));
    } // drawPort() //

    // takes the cell,cell xy position of a 1-cell gold and draws it
    public boolean drawGold(int cell_x, int cell_y)
    {
        int x,y; // drawGoldPos needs upper x,y coordinates //
        x = this.xCell2Pixel(cell_x);
        y = this.yCell2Pixel(cell_y);
        return(this.drawGoldPos(x,y));
    } // drawGold() //
    // overloaded 2-cell gold case..
    public boolean drawGold(int c1_x, int c1_y, int c2_x, int c2_y)
    {
        int x1, y1, x2, y2;
        x1 = this.xCell2Pixel(c1_x);
        x2 = this.xCell2Pixel(c2_x);
        y1 = this.yCell2Pixel(c1_y);
        y2 = this.yCell2Pixel(c2_y);
        return(this.drawGoldPos(x1,y1,x2,y2));
    }

    // draws the interior grid of the maze -- 1 pixel thick //
    public boolean drawGrid()
    {
        for (int i=1; i < this.height; i++) {
            this.drawHorizontalLine(0, (int)(this.width *
                                       this.cellsize *
                                       this.magnification),
                                    (int)(i *
                                    this.cellsize *
                                    this.magnification),
                                    1);
        } // end for() //
        for (int i=1; i < this.width; i++) {
            this.drawVerticalLine((int)(i * this.cellsize
                                        * this.magnification),
                                        0,(int)(this.height *
                                       this.cellsize *
                                       this.magnification),
                                    1);
        } // end for() //

        return(true);
    } // end drawGrid() //
    // draws the outside wall border of the maze //
    public boolean drawBorder()
    {
        this.drawHorizontalLine(0,
                                (int)(this.cellsize *
                                      this.width *
                                      this.magnification),
                                0,
                                this.wall_thickness);
        this.drawHorizontalLine(0,
                               (int)(this.cellsize *
                                     this.width *
                                     this.magnification),
                               (int) (this.cellsize *
                                      this.height *
                                      this.magnification),
                               this.wall_thickness);
        this.drawVerticalLine(0,
                              0,
                              (int) (this.cellsize *
                                     this.height *
                                     this.magnification),
                              this.wall_thickness);
        this.drawVerticalLine((int) (this.cellsize *
                                     this.width *
                                     this.magnification),
                              0,
                              (int) (this.cellsize *
                                     this.height *
                                     this.magnification),
                              this.wall_thickness);

        return(true);
    } // drawBorder() //

    // translates from cell,cell to x,y //
    public int xCell2Pixel(double cell)
    {
        return( (int)(cell * this.cellsize * this.magnification));
    } // xCell2Pixel //
    public int yCell2Pixel(double cell)
    {
        return( (int) ((this.height * this.cellsize * this.magnification) -
                   (this.cellsize * (cell+1) * this.magnification)));
    } // yCell2Pixel //
    // helper functions for just plain drawing... //

    public void drawSmoothRobot(SmoothRobotPose pose)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.green);
        int tl_x, tl_y;
        tl_x = this.xCell2Pixel(pose.x) + this.anchor_x;
        tl_y = this.yCell2Pixel(pose.y) + this.anchor_y;
        tl_x = tl_x + ((int)(this.wall_thickness * this.magnification * 2));
        tl_y = tl_y + ((int)(this.wall_thickness * this.magnification * 2));
        int diam = this.cellsize - (4 * this.wall_thickness);
        diam = ((int)(diam * this.magnification));
        int radius = (int)((float)diam / 2.0);

        g.drawOval(tl_x, tl_y, diam, diam);
        if (pose.num_golds >= 1) {
            this.drawGoldPos(tl_x-anchor_x,tl_y-anchor_y, false);
        } //endif
        if (pose.num_golds == 2) {
            this.drawGoldPos(tl_x-anchor_x+radius-
              ((int)(this.wall_thickness * this.magnification)),
              tl_y-anchor_y, false);
        } //endif
        g.setColor(Color.black);

        int cx = tl_x + radius;
        int cy = tl_y + radius;
        int rx = cx + (int) (radius * Math.cos(pose.angle));
        int ry = cy - (int) (radius * Math.sin(pose.angle));

        g.drawLine(cx, cy, rx, ry);

        g.dispose();
    }

    // draw a circle for a robot and a line to indicate its direction //
    // also draws zero, one or two golds inside the robot...
    public boolean drawRobotCircle(int x, int y, int dir, int golds, int robotnumber)
    {
        Graphics g = this.parent_frame.getGraphics();
        if ((robotnumber % 2) == 0)
             g.setColor(Color.green);
        else g.setColor(Color.green);
        int tl_x, tl_y;
        tl_x = x + this.anchor_x;
        tl_y = y + this.anchor_y;
        tl_x = tl_x + ((int)(this.wall_thickness * this.magnification * 2));
        tl_y = tl_y + ((int)(this.wall_thickness * this.magnification * 2));
        int diam = this.cellsize - (4 * this.wall_thickness);
        diam = ((int)(diam * this.magnification));
        int radius = (int)((float)diam / 2.0);

        g.fillOval(tl_x, tl_y, diam, diam);
        if (golds >= 1) {
            this.drawGoldPos(tl_x-anchor_x,tl_y-anchor_y, true);
        } //endif
        if (golds == 2) {
            this.drawGoldPos(tl_x-anchor_x+radius-
              ((int)(this.wall_thickness * this.magnification)),
              tl_y-anchor_y, true);
        } //endif

        g.dispose();
        return(true);
   } // drawRobotCircle()

   public boolean drawRobotLine(int x, int y, int dir, int golds, int robotnumber)
    {
        Graphics g = this.parent_frame.getGraphics();
        if ((robotnumber % 2) == 0)
             g.setColor(Color.green);
        else g.setColor(Color.green);
        int tl_x, tl_y;
        tl_x = x + this.anchor_x;
        tl_y = y + this.anchor_y;
        tl_x = tl_x + ((int)(this.wall_thickness * this.magnification * 2));
        tl_y = tl_y + ((int)(this.wall_thickness * this.magnification * 2));
        int diam = this.cellsize - (4 * this.wall_thickness);
        diam = ((int)(diam * this.magnification));
        int radius = (int)((float)diam / 2.0);

        /* g.fillOval(tl_x, tl_y, diam, diam); */
        /*
        if (golds >= 1) {
            this.drawGoldPos(tl_x-anchor_x,tl_y-anchor_y);
        } //endif
        if (golds == 2) {
            this.drawGoldPos(tl_x-anchor_x+radius-
              ((int)(this.wall_thickness * this.magnification)),
              tl_y-anchor_y);
        } //endif
        */

        g.setColor(Color.black);
        if (dir == 0) {
            g.drawLine(tl_x + radius, tl_y, tl_x + radius, tl_y + radius);
        }
        if (dir == 1) {
            g.drawLine(tl_x, tl_y + radius, tl_x + radius, tl_y + radius);
        }
        if (dir == 2) {
            g.drawLine(tl_x + radius, tl_y + radius,
                       tl_x + radius, tl_y + radius + radius);
        }
        if (dir == 3) {
            g.drawLine(tl_x + radius, tl_y + radius,
                       tl_x + radius + radius, tl_y + radius);
        }
        g.dispose();
        return(true);
    } // drawRobotLine() //

    public boolean drawRect(int x, int y, int width,
                            int height,
                            int thickness)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.red);
        for (int i=0; i<thickness; i++) {
            g.drawRect(this.anchor_x+x+i,
                       this.anchor_y+y+i,
                       width - (i*2), height - (i*2));
        } // end for //
        g.dispose();
        return(true);
    }

    //this takes tl pixel corner of one cell and draws the 1-cell gold
    public boolean drawPortPos(int x, int y)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.blue);
        int tl_x, tl_y;
        tl_x = x + this.anchor_x+ (int)(this.wall_thickness * this.magnification) - 2;
        tl_y = y + this.anchor_y + (int)(this.wall_thickness * this.magnification) - 2;
        g.drawRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification)+4,
                   (int)(this.wall_thickness * 2*this.magnification)+4);
        g.dispose();
        return(true);
    } // drawPortPos() //

    public boolean drawGoldPos(int x, int y)
    {
        return drawGoldPos(x, y, true);
    }

    //this takes tl pixel corner of one cell and draws the 1-cell gold
    public boolean drawGoldPos(int x, int y, boolean fill)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.orange);
        int tl_x, tl_y;
        tl_x = x + this.anchor_x+ (int)(this.wall_thickness * this.magnification);
        tl_y = y + this.anchor_y + (int)(this.wall_thickness * this.magnification);
        if(fill)
            g.fillRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification),
                       (int)(this.wall_thickness * 2*this.magnification));
        else
            g.drawRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification),
                       (int)(this.wall_thickness * 2*this.magnification));
        g.dispose();
        return(true);
    } // drawGold() //

    public boolean erasePortPos(int x, int y)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.white);
        int tl_x, tl_y;
        tl_x = xCell2Pixel(x) + this.anchor_x+ (int)(this.wall_thickness * this.magnification) - 2;
        tl_y = yCell2Pixel(y) + this.anchor_y + (int)(this.wall_thickness * this.magnification) - 2;
        g.fillRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification)+4,
                   (int)(this.wall_thickness * 2*this.magnification)+4);
        g.dispose();
        return(true);
    } // erasePortPos() //

    public boolean eraseGoldPos(int x, int y)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.white);
        int tl_x, tl_y;
        tl_x = xCell2Pixel(x) + this.anchor_x+ (int)(this.wall_thickness * this.magnification);
        tl_y = yCell2Pixel(y) + this.anchor_y + (int)(this.wall_thickness * this.magnification);
        g.fillRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification),
                   (int)(this.wall_thickness * 2*this.magnification));
        g.dispose();
        return(true);
    } // drawGold() //

    //takes the tl corner of two cells and draws a 2-cell covering gold
    public boolean drawGoldPos(int x1, int y1, int x2, int y2)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.orange);
        int tl_x, tl_y;
        if (x1 == x2) { // vertical pair of cells //
            tl_x = x1 + (int)(this.cellsize * this.magnification * 0.5);
            tl_x = tl_x - (int)(this.wall_thickness * this.magnification);
            tl_x = tl_x + this.anchor_x;
            tl_y = Math.max(y1,y2) - (int)(this.wall_thickness * this.magnification);
            tl_y = tl_y + this.anchor_y;
        } else if (y1 == y2) { // horizontal pair of cells //
            tl_y = y1 + (int)(this.cellsize * this.magnification*0.5);
            tl_y = tl_y - (int)(this.wall_thickness * this.magnification);
            tl_y = tl_y + this.anchor_y;
            tl_x = Math.max(x1,x2) - (int)(this.wall_thickness * this.magnification);
            tl_x = tl_x + this.anchor_x;
        } else {
            g.dispose();
            return(false);
        } //endelse#2
        g.fillRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification),
                   (int)(this.wall_thickness * 2*this.magnification));
        g.dispose();
        return(true);

    } // drawGoldPos() //

//takes the tl corner of two cells and draws a 2-cell covering gold
    public boolean eraseGoldPos(int x1, int y1, int x2, int y2)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.white);
        int tl_x, tl_y;
        if (x1 == x2) { // vertical pair of cells //
            tl_x = x1 + (int)(this.cellsize * this.magnification * 0.5);
            tl_x = tl_x - (int)(this.wall_thickness * this.magnification);
            tl_x = tl_x + this.anchor_x;
            tl_y = Math.max(y1,y2) - (int)(this.wall_thickness * this.magnification);
            tl_y = tl_y + this.anchor_y;
        } else if (y1 == y2) { // horizontal pair of cells //
            tl_y = y1 + (int)(this.cellsize * this.magnification*0.5);
            tl_y = tl_y - (int)(this.wall_thickness * this.magnification);
            tl_y = tl_y + this.anchor_y;
            tl_x = Math.max(x1,x2) - (int)(this.wall_thickness * this.magnification);
            tl_x = tl_x + this.anchor_x;
        } else {
            g.dispose();
            return(false);
        } //endelse#2
        g.fillRect(tl_x, tl_y, (int)(this.wall_thickness * 2*this.magnification),
                   (int)(this.wall_thickness * 2*this.magnification));
        g.dispose();
        return(true);

    } // drawGoldPos() //

    // this draws a line and does thickness symmetrically //
    public boolean drawHorizontalLine(int x1, int x2,
                                      int y, int thickness)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.red);
        int start_y = y - (int)((float)(thickness)/2.0);
        for (int i=0; i<thickness; i++) {
            g.drawLine(this.anchor_x+x1,
                       this.anchor_y+start_y+i,
                       this.anchor_x+x2,
                       this.anchor_y+start_y+i);
        } // end for() //
        g.dispose();
        return(true);
    } // drawHorizontalLine() //

    // this draws a line and does thickness symmetrically //
    public boolean drawVerticalLine(int x, int y1,
                                      int y2, int thickness)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.red);
        int start_x = x - (int)((float)(thickness)/2.0);
        for (int i=0; i<thickness; i++) {
            g.drawLine(this.anchor_x+start_x+i,
                       this.anchor_y+y1,
                       this.anchor_x+start_x+i,
                       this.anchor_y+y2);
        } // end for() //
        g.dispose();
        return(true);
    } // drawVerticalLine() //

    // this draws a line and does thickness symmetrically //
    public boolean eraseHorizontalLine(int x1, int x2,
                                      int y, int thickness)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.white);
        int start_y = y - (int)((float)(thickness)/2.0);
        for (int i=0; i<thickness; i++) {
            g.drawLine(this.anchor_x+x1,
                       this.anchor_y+start_y+i,
                       this.anchor_x+x2,
                       this.anchor_y+start_y+i);
        } // end for() //
        g.dispose();
        return(true);
    } // drawHorizontalLine() //

    // this draws a line and does thickness symmetrically //
    public boolean eraseVerticalLine(int x, int y1,
                                      int y2, int thickness)
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.white);
        int start_x = x - (int)((float)(thickness)/2.0);
        for (int i=0; i<thickness; i++) {
            g.drawLine(this.anchor_x+start_x+i,
                       this.anchor_y+y1,
                       this.anchor_x+start_x+i,
                       this.anchor_y+y2);
        } // end for() //
        g.dispose();
        return(true);
    } // drawVerticalLine() //

    // this takes the entire maze off the screen by drawing a filled white rect
    public boolean wipeMaze()
    {
        Graphics g = this.parent_frame.getGraphics();
        g.setColor(Color.white);
        g.fillRect(this.anchor_x-
                    (int)(this.magnification *
                     this.wall_thickness),
                   this.anchor_y-
                    (int)(this.magnification *
                     this.wall_thickness),
                   (this.width * this.cellsize)+
                   (2 * (int)(this.magnification *
                         this.wall_thickness)),
                   (this.height * this.cellsize)+
                   (2 * (int)(this.magnification *
                         this.wall_thickness)));
        g.dispose();
        return(true);
    } // wipeMaze()

    public void refresh ()
    {
        this.wipeMaze();
        this.drawBorder();
        this.drawGrid();
        this.drawWalls();
        this.drawRobots();
        this.drawPorts();
        this.drawGolds();
    } // refresh() //

    public void update ()
    {   this.refresh();
    } // update() //

} // end MazeGraphics //
